#include <msp/debug/demangle.h>
#include <msp/gl/matrix.h>
#include <msp/gl/meshbuilder.h>
#include "graphic.h"
#include "list.h"
#include "part.h"
#include "style.h"
#include "text.h"
#include "vslider.h"

using namespace std;

namespace Msp {
namespace GLtk {

incompatible_data::incompatible_data(const type_info &ti):
	logic_error("expected "+Debug::demangle(ti.name()))
{ }


List::List():
	data(new BasicListData<string>),
	own_data(true)
{
	init();
}

List::List(ListData &d):
	data(&d),
	own_data(false)
{
	init();
}

void List::init()
{
	item_factory = 0;
	sel_index = -1;
	first = 0;
	max_scroll = 0;
	view_size = 5;

	observer = new DataObserver(*this);

	add(slider);
	slider.set_step(1);
	slider.signal_value_changed.connect(sigc::mem_fun(this, &List::slider_value_changed));
}

List::~List()
{
	delete item_factory;
	delete observer;
	if(own_data)
		delete data;
}

void List::autosize_special(const Part &part, Geometry &ageom) const
{
	if(part.get_name()=="items")
	{
		const Sides &margin = part.get_margin();

		unsigned max_w = 0;
		unsigned total_h = 0;
		for(unsigned i=0; i<items.size(); ++i)
		{
			Geometry igeom;
			items[i]->autosize(igeom);
			max_w = max(max_w, igeom.w);
			if(view_size==0 || i<view_size)
				total_h += igeom.h;
		}

		if(!items.empty() && items.size()<view_size)
			total_h = total_h*view_size/items.size();

		ageom.w = max(ageom.w, max_w+margin.left+margin.right);
		ageom.h = max(ageom.h, total_h+margin.top+margin.bottom);
	}
	else if(part.get_name()=="slider")
		autosize_child(slider, part, ageom);
}

void List::set_data(ListData &d)
{
	if(item_factory)
		item_factory->set_data(d);

	delete observer;
	if(own_data)
		delete data;

	data = &d;
	own_data = false;
	observer = new DataObserver(*this);

	for(vector<Item *>::iterator i=items.begin(); i!=items.end(); ++i)
		delete *i;
	items.clear();
	unsigned n_items = data->size();
	for(unsigned i=0; i<n_items; ++i)
	{
		Item *item = create_item(i);
		items.push_back(item);
	}

	items_changed();
}

void List::items_changed()
{
	signal_autosize_changed.emit();
	rebuild();
}

List::Item *List::create_item(unsigned index)
{
	Item *item = 0;	
	if(item_factory)
		item = item_factory->create_item(index);
	else
		item = new BasicItem(data->get_string(index));
	item->signal_autosize_changed.connect(sigc::bind(sigc::mem_fun(this, &List::item_autosize_changed), item));
	add(*item);
	return item;
}

void List::set_view_size(unsigned s)
{
	view_size = s;
	signal_autosize_changed.emit();
}

void List::set_view_all()
{
	set_view_size(0);
}

void List::set_selected_index(int i)
{
	if(i>=static_cast<int>(data->size()))
		throw out_of_range("List::set_selected_index");

	if(i==sel_index)
		return;

	if(sel_index>=0)
		items[sel_index]->set_active(false);
	if(i<0)
		sel_index = -1;
	else
	{
		sel_index = i;
		items[sel_index]->set_active(true);
		signal_item_selected.emit(sel_index);
	}
}

void List::rebuild_special(const Part &part)
{
	if(part.get_name()=="slider")
		reposition_child(slider, part);
	else if(part.get_name()=="items")
	{
		const Sides &margin = part.get_margin();
		unsigned w = geom.w-min(geom.w, margin.left+margin.right);
		unsigned y = geom.h-min(geom.h, margin.top);
		for(unsigned i=0; i<items.size(); ++i)
		{
			if(i<first || !y)
				items[i]->set_visible(false);
			else
			{
				Geometry igeom = items[i]->get_geometry();
				if(igeom.h+margin.bottom<=y)
				{
					items[i]->set_visible(true);
					y -= igeom.h;
					igeom.x = margin.left;
					igeom.y = y;
					igeom.w = w;
					items[i]->set_geometry(igeom);
				}
				else
				{
					items[i]->set_visible(false);
					y = 0;
				}
			}
		}

		check_view_range();
	}

	Widget::rebuild_special(part);
}

void List::render_special(const Part &part, GL::Renderer &renderer) const
{
	if(part.get_name()=="items")
	{
		for(unsigned i=first; (i<items.size() && items[i]->is_visible()); ++i)
			items[i]->render(renderer);
	}
	else if(part.get_name()=="slider")
		slider.render(renderer);
}

void List::button_press(int x, int y, unsigned btn)
{
	Container::button_press(x, y, btn);
	if(click_focus && btn==1)
	{
		for(unsigned i=first; (i<items.size() && items[i]->is_visible()); ++i)
			if(click_focus==items[i])
			{
				set_selected_index(i);
				break;
			}
	}
}

void List::item_autosize_changed(Item *item)
{
	item->autosize();
	signal_autosize_changed.emit();
	rebuild();
}

void List::check_view_range()
{
	if(!style)
		return;

	unsigned h = geom.h;
	if(const Part *items_part = style->get_part("items"))
	{
		const Sides &margin = items_part->get_margin();
		h -= margin.top+margin.bottom;
	}

	max_scroll = items.size();
	for(unsigned i=items.size(); i-->0; )
	{
		unsigned ih = items[i]->get_geometry().h;
		if(ih<=h)
		{
			h -= ih;
			--max_scroll;
		}
	}

	if(first>max_scroll)
		first = max_scroll;

	slider.set_range(0, max_scroll);
	slider.set_value(max_scroll-first);
}

void List::slider_value_changed(double value)
{
	if(max_scroll>0)
	{
		first = max_scroll-static_cast<unsigned>(value);
		rebuild();
	}
}


List::DataObserver::DataObserver(List &l):
	list(l)
{
	list.data->signal_item_added.connect(sigc::mem_fun(this, &DataObserver::item_added));
	list.data->signal_item_removed.connect(sigc::mem_fun(this, &DataObserver::item_removed));
	list.data->signal_cleared.connect(sigc::mem_fun(this, &DataObserver::cleared));
	list.data->signal_refresh_item.connect(sigc::mem_fun(this, &DataObserver::refresh_item));
}

void List::DataObserver::item_added(unsigned i)
{
	if(list.sel_index>=static_cast<int>(i))
		++list.sel_index;

	Item *item = list.create_item(i);
	list.items.insert(list.items.begin()+i, item);
	list.items_changed();
}

void List::DataObserver::item_removed(unsigned i)
{
	if(list.sel_index>static_cast<int>(i))
		--list.sel_index;
	else if(list.sel_index==static_cast<int>(i))
		list.sel_index = -1;

	delete list.items[i];
	list.items.erase(list.items.begin()+i);
	list.items_changed();
}

void List::DataObserver::cleared()
{
	list.sel_index = -1;
	for(vector<Item *>::iterator i=list.items.begin(); i!=list.items.end(); ++i)
		delete *i;
	list.items.clear();
	list.items_changed();
}

void List::DataObserver::refresh_item(unsigned i)
{
	delete list.items[i];
	list.items[i] = list.create_item(i);
	list.items_changed();
}


void List::Item::autosize_special(const Part &part, Geometry &ageom) const
{
	if(part.get_name()=="children")
	{
		const Sides &margin = part.get_margin();
		for(list<Child *>::const_iterator i=children.begin(); i!=children.end(); ++i)
		{
			Geometry cgeom;
			(*i)->widget->autosize(cgeom);
			ageom.w = max(ageom.w, cgeom.x+cgeom.w+margin.right);
			ageom.h = max(ageom.h, cgeom.y+cgeom.h+margin.top);
		}
	}
}

void List::Item::set_active(bool a)
{
	set_state(ACTIVE, (a ? ACTIVE : NORMAL));
}

void List::Item::render_special(const Part &part, GL::Renderer &renderer) const
{
	if(part.get_name()=="children")
	{
		for(list<Child *>::const_iterator i=children.begin(); i!=children.end(); ++i)
			(*i)->widget->render(renderer);
	}
}


void List::MultiColumnItem::check_widths(vector<unsigned> &widths) const
{
	if(widths.size()<children.size())
		widths.resize(children.size(), 0);

	unsigned n = 0;
	for(list<Child *>::const_iterator i=children.begin(); i!=children.end(); ++i, ++n)
	{
		Geometry cgeom;
		(*i)->widget->autosize(cgeom);
		// TODO invent a better way to specify spacings
		widths[n] = max(widths[n], cgeom.w+8);
	}
}

void List::MultiColumnItem::set_widths(const vector<unsigned> &widths)
{
	if(!style)
		return;

	const Part *part = style->get_part("children");
	if(!part)
		return;

	const Sides &margin = part->get_margin();
	int x = margin.left;
	unsigned n = 0;
	for(list<Child *>::const_iterator i=children.begin(); i!=children.end(); ++i, ++n)
	{
		(*i)->widget->set_position(x, margin.bottom);
		x += widths[n];
	}
}

void List::MultiColumnItem::on_style_change()
{
	if(!style)
		return;

	for(std::list<Child *>::const_iterator i=children.begin(); i!=children.end(); ++i)
		(*i)->widget->autosize();

	vector<unsigned> widths;
	List *list = static_cast<List *>(parent);
	for(vector<Item *>::const_iterator i=list->items.begin(); i!=list->items.end(); ++i)
		if(*i!=this)
			if(MultiColumnItem *mci = dynamic_cast<MultiColumnItem *>(*i))
				mci->check_widths(widths);

	vector<unsigned> self_widths(widths);
	check_widths(self_widths);
	bool update_all = false;
	for(unsigned i=0; (!update_all && i<widths.size() && i<self_widths.size()); ++i)
		update_all = self_widths[i]>widths[i];

	if(update_all)
	{
		for(vector<Item *>::const_iterator i=list->items.begin(); i!=list->items.end(); ++i)
			if(MultiColumnItem *mci = dynamic_cast<MultiColumnItem *>(*i))
				mci->set_widths(self_widths);
	}
	else
		set_widths(self_widths);
}


List::BasicItem::BasicItem(const string &text):
	label(text)
{
	add(label);
}

void List::BasicItem::on_style_change()
{
	if(!style)
		return;

	label.autosize();
	if(const Part *part = style->get_part("children"))
	{
		const Sides &margin = part->get_margin();
		label.set_position(margin.left, margin.bottom);
	}
}


List::Loader::Loader(List &l):
	DataFile::DerivedObjectLoader<List, Widget::Loader>(l)
{
	add("item", &Loader::item);
	add("view_size", &List::view_size);
}

void List::Loader::item(const string &v)
{
	dynamic_cast<BasicListData<string> &>(*obj.data).append(v);
}

} // namespace GLtk
} // namespace Msp
